package FCC::App::Actions::Index;
$VERSION = 1.00;
use strict;
use warnings;
use utf8;
use base qw(FCC::App::Actions::_Super);
use File::Which;

sub dispatch {
    my ($self) = @_;

    my $body = $self->output_template(
        {
            # サーバー情報
            server_info => $self->get_server_info(),

            # インストールされている Perl のパスとバージョン
            perl_list => $self->get_perl_list(),

            # CGI 実行環境
            exec_env => $self->get_exec_env(),

            # 各種コマンドパス
            cmdpath_list => $self->get_cmdpath_list(),

            # DNS逆引き
            dns_reverse_lookup => $self->get_dns_reverse_lookup(),

            # 環境変数
            envvar_list => $self->get_envvar_list(),

            # Perl モジュール検索パス
            libpath_list => $self->get_libpath_list(),

            # Perl モジュール一覧
            module_list => $self->get_module_list()
        }
    );
    $self->print_html($body);
}

# サーバー情報
sub get_server_info {
    my ($self) = @_;
    my $info = {
        machine  => `uname -m`,
        nodename => `uname -n`,
        release  => `uname -r`,
        sysname  => `uname -s`,
        version  => `uname -v`
    };
    return $info;
}

# インストールされている Perl のパスとバージョン
sub get_perl_list {
    my ($self) = @_;
    my @perl_path_list = File::Which::where 'perl';
    my @perl_list;
    for my $path (@perl_path_list) {
        my $version  = '';
        my $selected = 0;
        my $symlink  = '';
        if ( -l $path ) {
            $symlink = $self->read_symlink_target($path);
        }

        eval {
            $version = `${path} -e 'print \$\^V' 2>/dev/null`;
            unless ($version) {
                my $res = `${path} -v`;
                ($version) =
                  $res =~ /[\(\s](v\d+\.\d+\.\d+)[\)\s]/;    # 例: 'v5.30.3'
            }
            if ( $version eq $^V ) {
                $selected = 1;
            }
        };

        push(
            @perl_list,
            {
                path     => $path,
                version  => $version,
                selected => $selected,
                symlink  => $symlink
            }
        );
    }
    return \@perl_list;
}

sub read_symlink_target {
    my ( $self, $path ) = @_;
    unless ( -l $path ) {
        return $path;
    }
    my $linkpath = readlink($path);
    if ( -l $linkpath ) {
        return $self->read_symlink_target($linkpath);
    }
    else {
        return $linkpath;
    }
}

# CGI 実行環境
sub get_exec_env {
    my ($self)    = @_;
    my $owneruid  = ( stat( $ENV{SCRIPT_FILENAME} ) )[4];
    my $ownername = getpwuid($owneruid);
    my $execuid   = $<;
    my $execname  = getpwuid($execuid);
    my $suexec    = ( $owneruid == $execuid ) ? 1 : 0;

    my $env = {
        owneruid  => $owneruid,
        ownername => $ownername,
        execuid   => $execuid,
        execname  => $execname,
        suexec    => $suexec
    };
    return $env;
}

# 各種コマンドパス
sub get_cmdpath_list {
    my ($self) = @_;
    my @cmd_list = (
        'sendmail', 'php', 'python', 'python3',
        'ruby',     'nkf', 'gzip',   'gunzip'
    );
    my @search_path_list = (
        '/bin',      '/usr/bin', '/usr/local/bin', '/sbin',
        '/usr/sbin', '/usr/lib', '/usr/local/lib'
    );
    my @cmdpath_list;

    for my $cmd (@cmd_list) {
        my @path_list = File::Which::where $cmd;
        unless (@path_list) {
            for my $dpath (@search_path_list) {
                my $fpath = "${dpath}/${cmd}";
                if ( -e $fpath && -x $fpath ) {
                    push( @path_list, $fpath );
                }
            }
        }

        push(
            @cmdpath_list,
            {
                command   => $cmd,
                path_list => \@path_list
            }
        );
    }
    return \@cmdpath_list;
}

# DNS逆引き
sub get_dns_reverse_lookup {
    my ($self)    = @_;
    my $env_rhost = $ENV{REMOTE_HOST};
    my $rev_rhost = $self->get_remote_host();

    my $data = {
        env_rhost    => $env_rhost,
        env_rhost_ok => ( $env_rhost =~ /[a-zA-Z]/ ) ? 1 : 0,
        rev_rhost    => $rev_rhost,
        rev_rhost_ok => ( $rev_rhost =~ /[a-zA-Z]/ ) ? 1 : 0
    };

    return $data;
}

sub get_remote_host {
    my ($self)        = @_;
    my @addr          = split( /\./, $ENV{REMOTE_ADDR} );
    my $packed_addr   = pack( "C4", $addr[0], $addr[1], $addr[2], $addr[3] );
    my ($remote_host) = gethostbyaddr( $packed_addr, 2 );
    return $remote_host;
}

# 環境変数
sub get_envvar_list {
    my ($self) = @_;
    my @envvar_list;
    for my $name ( sort { $a cmp $b } keys %ENV ) {
        push(
            @envvar_list,
            {
                name  => $name,
                value => $ENV{$name}
            }
        );
    }
    return \@envvar_list;
}

# Perl モジュール検索パス
sub get_libpath_list {
    my ($self) = @_;
    my $this_path = $self->{conf}->{CGI_DIR};
    my @list;
    for my $path (@INC) {
        if ( $path eq '.' ) {
            next;
        }
        unless ( $path =~ /^\Q${this_path}\E/ ) {
            push( @list, $path );
        }
    }
    return \@list;
}

# Perl モジュール一覧
sub get_module_list {
    my ($self) = @_;
    my $search_path_list = $self->get_libpath_list();

    # 重複したパスを排除
    my $paths = {};
    for my $path ( @{$search_path_list} ) {
        $paths->{$path} = 1;
    }

    # 包含関係にあるパスを排除 (子となるパスを排除)
    my @path_list;
    my $prev_path = '';
    for my $path ( sort { $a cmp $b } keys %{$paths} ) {
        if ( $prev_path && $path =~ /^\Q${prev_path}\E/ ) {
            next;
        }
        push( @path_list, $path );
        $prev_path = $path;
    }

    # モジュールを検索
    my @module_list;
    for my $path (@path_list) {
        my $mlist = $self->find_modules($path);
        for my $module ( @{$mlist} ) {
            push( @module_list, $module );
        }
    }

    # モジュール名でソート
    @module_list = sort { $a->{name} cmp $b->{name} } @module_list;
    return \@module_list;
}

sub find_modules {
    my ( $self, $path ) = @_;
    my $module_list = [];

    unless ( -d $path ) {
        return $module_list;
    }

    opendir my $dh, $path;
    my @files = readdir($dh);
    closedir $dh;

    for my $file (@files) {
        if ( $file =~ /^\./ ) { next; }
        my $dpath = "${path}/${file}";
        if ( -d $dpath ) {
            my $mlist = $self->find_modules($dpath);
            for my $module ( @{$mlist} ) {
                push( @{$module_list}, $module );
            }
        }
        else {
            unless ( $file =~ /\.pm$/ ) {
                next;
            }
            open my $fh, '<', $dpath or next;
            while ( my $line = <$fh> ) {
                if ( $line =~ /^\s*package\s+(\S+)\s*;/ ) {
                    my $mname = $1;
                    push(
                        @{$module_list},
                        {
                            name => $1,
                            path => $dpath
                        }
                    );
                    last;
                }
            }
            close $fh;
        }
    }
    return $module_list;
}

1;
